"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.AsarPackager = void 0;
const asar_1 = require("@electron/asar");
const builder_util_1 = require("builder-util");
const builder_util_runtime_1 = require("builder-util-runtime");
const fs_1 = require("builder-util/out/fs");
const fsNode = require("fs");
const fs = require("fs-extra");
const path = require("path");
const tempFile = require("temp-file");
const appFileCopier_1 = require("../util/appFileCopier");
const unpackDetector_1 = require("./unpackDetector");
/** @internal */
class AsarPackager {
    constructor(config) {
        this.config = config;
        this.tmpDir = new tempFile.TmpDir();
        this.outFile = path.join(config.resourcePath, `app.asar`);
    }
    async pack(fileSets, _packager) {
        this.rootForAppFilesWithoutAsar = await this.tmpDir.getTempDir({ prefix: "asar-app" });
        const cancellationToken = new builder_util_runtime_1.CancellationToken();
        cancellationToken.on("cancel", () => this.tmpDir.cleanupSync());
        const orderedFileSets = [
            // Write dependencies first to minimize offset changes to asar header
            ...fileSets.slice(1),
            // Finish with the app files that change most often
            fileSets[0],
        ].map(orderFileSet);
        const { unpackedPaths, copiedFiles } = await this.detectAndCopy(orderedFileSets, cancellationToken);
        const unpackGlob = unpackedPaths.length > 1 ? `{${unpackedPaths.join(",")}}` : unpackedPaths.pop();
        let ordering = this.config.options.ordering || undefined;
        if (!ordering) {
            // `copiedFiles` are already ordered due to `orderedFileSets` input, so we just map to their relative paths (via substring) within the asar.
            const filesSorted = copiedFiles.map(file => file.substring(this.rootForAppFilesWithoutAsar.length));
            ordering = await this.tmpDir.getTempFile({ prefix: "asar-ordering", suffix: ".txt" });
            await fs.writeFile(ordering, filesSorted.join("\n"));
        }
        const options = {
            unpack: unpackGlob,
            unpackDir: unpackGlob,
            ordering,
            dot: true,
        };
        // override logger temporarily to clean up console (electron/asar does some internal logging that blogs up the default electron-builder logs)
        const consoleLogger = console.log;
        console.log = (...args) => {
            if (args[0] === "Ordering file has 100% coverage.") {
                return; // no need to log, this means our ordering logic is working correctly
            }
            builder_util_1.log.info({ args }, "logging @electron/asar");
        };
        await (0, asar_1.createPackageWithOptions)(this.rootForAppFilesWithoutAsar, this.outFile, options);
        console.log = consoleLogger;
        await this.tmpDir.cleanup();
    }
    async detectAndCopy(fileSets, cancellationToken) {
        var _a;
        const taskManager = new builder_util_1.AsyncTaskManager(cancellationToken);
        const unpackedPaths = new Set();
        const copiedFiles = new Set();
        const matchUnpacker = (file, dest, stat) => {
            var _a, _b;
            if ((_b = (_a = this.config).unpackPattern) === null || _b === void 0 ? void 0 : _b.call(_a, file, stat)) {
                builder_util_1.log.debug({ file }, "unpacking");
                unpackedPaths.add(dest);
                return;
            }
        };
        const writeFileOrSymlink = async (transformedData, source, destination, stat) => {
            copiedFiles.add(destination);
            // If transformed data, skip symlink logic
            if (transformedData) {
                return this.copyFileOrData(transformedData, source, destination, stat);
            }
            const realPathFile = await fs.realpath(source);
            if (source === realPathFile) {
                return this.copyFileOrData(undefined, source, destination, stat);
            }
            const realPathRelative = path.relative(this.config.appDir, realPathFile);
            const symlinkTarget = path.resolve(this.rootForAppFilesWithoutAsar, realPathRelative);
            const isOutsidePackage = realPathRelative.startsWith("..");
            if (isOutsidePackage) {
                builder_util_1.log.error({ source: builder_util_1.log.filePath(source), realPathFile: builder_util_1.log.filePath(realPathFile) }, `unable to copy, file is symlinked outside the package`);
                throw new Error(`Cannot copy file (${path.basename(source)}) symlinked to file (${path.basename(realPathFile)}) outside the package as that violates asar security integrity`);
            }
            await this.copyFileOrData(undefined, source, symlinkTarget, stat);
            const target = path.relative(path.dirname(destination), symlinkTarget);
            fsNode.symlinkSync(target, destination);
            copiedFiles.add(symlinkTarget);
        };
        for await (const fileSet of fileSets) {
            if (this.config.options.smartUnpack !== false) {
                (0, unpackDetector_1.detectUnpackedDirs)(fileSet, unpackedPaths, this.config.defaultDestination);
            }
            for (let i = 0; i < fileSet.files.length; i++) {
                const file = fileSet.files[i];
                const transformedData = (_a = fileSet.transformedFiles) === null || _a === void 0 ? void 0 : _a.get(i);
                const metadata = fileSet.metadata.get(file) || (await fs.lstat(file));
                const relative = path.relative(this.config.defaultDestination, (0, appFileCopier_1.getDestinationPath)(file, fileSet));
                const dest = path.resolve(this.rootForAppFilesWithoutAsar, relative);
                matchUnpacker(file, dest, metadata);
                taskManager.addTask(writeFileOrSymlink(transformedData, file, dest, metadata));
                if (taskManager.tasks.length > fs_1.MAX_FILE_REQUESTS) {
                    await taskManager.awaitTasks();
                }
            }
        }
        await taskManager.awaitTasks();
        return {
            unpackedPaths: Array.from(unpackedPaths),
            copiedFiles: Array.from(copiedFiles),
        };
    }
    async copyFileOrData(data, source, destination, stat) {
        await fs.mkdir(path.dirname(destination), { recursive: true });
        if (data) {
            await fs.writeFile(destination, data, { mode: stat.mode });
        }
        else {
            await fs.copyFile(source, destination);
        }
    }
}
exports.AsarPackager = AsarPackager;
function orderFileSet(fileSet) {
    const sortedFileEntries = Array.from(fileSet.files.entries());
    sortedFileEntries.sort(([, a], [, b]) => {
        if (a === b) {
            return 0;
        }
        // Place addons last because their signature changes per build
        const isAAddon = a.endsWith(".node");
        const isBAddon = b.endsWith(".node");
        if (isAAddon && !isBAddon) {
            return 1;
        }
        if (isBAddon && !isAAddon) {
            return -1;
        }
        // Otherwise order by name
        return a < b ? -1 : 1;
    });
    let transformedFiles;
    if (fileSet.transformedFiles) {
        transformedFiles = new Map();
        const indexMap = new Map();
        for (const [newIndex, [oldIndex]] of sortedFileEntries.entries()) {
            indexMap.set(oldIndex, newIndex);
        }
        for (const [oldIndex, value] of fileSet.transformedFiles) {
            const newIndex = indexMap.get(oldIndex);
            if (newIndex === undefined) {
                const file = fileSet.files[oldIndex];
                throw new Error(`Internal error: ${file} was lost while ordering asar`);
            }
            transformedFiles.set(newIndex, value);
        }
    }
    const { src, destination, metadata } = fileSet;
    return {
        src,
        destination,
        metadata,
        files: sortedFileEntries.map(([, file]) => file),
        transformedFiles,
    };
}
//# sourceMappingURL=asarUtil.js.map